"""SSL Contexts"""

import logging
import ssl
from typing import (
    AnyStr,
    Callable,
    Iterable,
    Optional
)

from .constants import DEFAULT_PROTOCOLS

LOGGER = logging.getLogger(__name__)

DEFAULT_CIPHERS = (
    "ECDHE+AESGCM",
    "ECDHE+CHACHA20",
    "DHE+AESGCM",
    "DHE+CHACHA20",
    "ECDH+AESGCM",
    "DH+AESGCM",
    "ECDH+AES",
    "DH+AES",
    "RSA+AESGCM",
    "RSA+AES",
    "!aNULL",
    "!eNULL",
    "!MD5",
    "!DSS",
)
DEFAULT_OPTIONS = (
    ssl.OP_NO_SSLv2,
    ssl.OP_NO_SSLv3,
    ssl.OP_NO_TLSv1,
    ssl.OP_NO_TLSv1_1,
    ssl.OP_NO_COMPRESSION
)


def create_ssl_context(
        cafile: Optional[str],
        capath: Optional[str],
        cadata: Optional[str],
        *,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS
) -> ssl.SSLContext:
    """Create an ssl context suitable for https

    Args:
        cafile (Optional[str]): The path of a file of concatenated CA
            certificates in PEM format.
        capath (Optional[str]): The path to a directory containing CA
            certificates in PEM format.
        cadata (Optional[str]): The data for a PEM encoded certificate.
        protocols (Iterable[str], optional): The supported protocols.
            Defaults to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The supported ciphers.
            Defaults to DEFAULT_CIPHERS.
        options (Iterable[str], optional): The SSLContext options.
            Defaults to DEFAULT_OPTIONS.

    Returns:
        ssl.SSLContext: An ssl context
    """
    ctx = ssl.create_default_context(
        purpose=ssl.Purpose.SERVER_AUTH,
        cafile=cafile,
        capath=capath,
        cadata=cadata
    )
    for option in options:
        ctx.options |= option
    ctx.set_ciphers(':'.join(ciphers))
    ctx.set_alpn_protocols(list(protocols))
    try:
        ctx.set_npn_protocols(list(protocols))
    except NotImplementedError:
        LOGGER.debug("Can't set npn protocols")
    return ctx


def create_ssl_context_with_cert_chain(
        certfile: str,
        keyfile: str,
        password: Optional[Callable[[], AnyStr]] = None,
        verify_mode: int = ssl.CERT_REQUIRED,
        check_hostname: bool = True,
        *,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS
) -> ssl.SSLContext:
    """Create an ssl context with load_cert_chain

    Args:
        certfile (str): The path to a certificate file.
        keyfile (str): The path to a key file.
        password (Optional[Callable[[], AnyStr]], optional): A function to get
            the password. Defaults to None.
        verify_mode (int, optional): The verify mode. Defaults to
            ssl.CERT_REQUIRED.
        check_hostname (bool, optional): Whether the hostname should be checked.
            Defaults to True.
        protocols (Iterable[str], optional): The alpn protocols.
            Defaults to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The ciphers.
            Defaults to DEFAULT_CIPHERS.
        options (Iterable[int], optional): The SSLContext options.
            Defaults to DEFAULT_OPTIONS.

    Returns:
        ssl.SSLContext: [description]
    """
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
    for option in options:
        ssl_context.options |= option
    ssl_context.set_ciphers(':'.join(ciphers))
    ssl_context.set_alpn_protocols(list(protocols))

    ssl_context.verify_mode = verify_mode
    ssl_context.check_hostname = check_hostname

    ssl_context.load_cert_chain(
        certfile=certfile,
        keyfile=keyfile,
        password=password  # type: ignore
    )

    return ssl_context



"""Requester"""

import urllib.parse
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    List,
    Mapping,
    Optional
)
from baretypes import Header
from bareutils.compression import compression_reader_adapter
import bareutils.header as header

from .acgi import ReceiveCallable, SendCallable
from .constants import USER_AGENT, Decompressors
from .utils import NullIter


def _enrich_headers(
        url: urllib.parse.ParseResult,
        headers: Optional[List[Header]],
        content: Optional[AsyncIterable[bytes]]
) -> List[Header]:
    headers = [] if not headers else list(headers)
    if not header.find(b'user-agent', headers):
        headers.append((b'user-agent', USER_AGENT))
    if not header.find(b'host', headers):
        headers.append((b'host', url.netloc.encode('ascii')))
    if content and not (
            header.find(b'content-length', headers)
            or header.find(b'transfer-encoding', headers)
    ):
        headers.append((b'transfer-encoding', b'chunked'))
    return headers


class RequestHandlerInstance:
    """The request handler"""

    def __init__(
            self,
            url: urllib.parse.ParseResult,
            method: str,
            headers: Optional[List[Header]],
            content: Optional[AsyncIterable[bytes]],
            send: SendCallable,
            receive: ReceiveCallable,
            decompressors: Decompressors
    ) -> None:
        """Initialise the request handler instance

        Args:
            url (urllib.parse.ParseResult): The parsed url
            method (str): The request method
            headers (Optional[List[Header]]): The request headers
            content (Optional[AsyncIterable[bytes]]): The content
            send (SendCallable): The function to sed the data
            receive (ReceiveCallable): The function to receive the data
            decompressors (Decompressors): The available
                decompressors
        """
        self.url = url
        self.method = method
        self.headers = _enrich_headers(url, headers, content)
        self.content = content
        self.send = send
        self.receive = receive
        self.decompressors = decompressors

    async def process(self) -> Mapping[str, Any]:
        """Process the request

        Returns:
            Mapping[str, Any]: The response message.
        """
        await self._process_request()
        return await self._process_response()

    async def _process_request(self) -> None:
        content_list: List[bytes] = []
        content_iter: AsyncIterator[bytes] = (
            NullIter() if self.content is None else
            self.content.__aiter__()
        )
        try:
            while len(content_list) < 2:
                body = await content_iter.__anext__()
                content_list.append(body)
        except StopAsyncIteration:
            pass

        body = content_list.pop(0) if content_list else b''
        more_body = len(content_list) > 0

        message: Mapping[str, Any] = {
            'type': 'http.request',
            'url': self.url,
            'method': self.method,
            'headers': self.headers,
            'body': body,
            'more_body': more_body
        }
        await self.send(message)

        connection = await self.receive()

        stream_id: Optional[int] = connection['stream_id']

        while more_body:
            try:
                body = await content_iter.__anext__()
                content_list.append(body)
            except StopAsyncIteration:
                pass

            body = content_list.pop(0) if content_list else b''
            more_body = len(content_list) > 0

            message = {
                'type': 'http.request.body',
                'body': body,
                'more_body': more_body,
                'stream_id': stream_id
            }
            await self.send(message)

    async def _process_response(self) -> Mapping[str, Any]:
        response = dict(await self.receive())

        response['body'] = self._make_body_reader(
            response['headers']
        ) if response.get('more_body', False) else None

        return response

    def _make_body_reader(
            self,
            headers: List[Header]
    ) -> AsyncIterable[bytes]:
        reader = self._body_reader()
        content_encoding = header.content_encoding(headers)
        if content_encoding:
            for encoding in content_encoding:
                if encoding in self.decompressors:
                    decompressor = self.decompressors[encoding]
                    return compression_reader_adapter(reader, decompressor())
        return reader

    async def _body_reader(self) -> AsyncIterator[bytes]:
        more_body = True
        while more_body:
            message = await self.receive()
            yield message.get('body', b'')
            more_body = message.get('more_body', False)

    async def close(self) -> None:
        """Close the request"""
        await self.send({
            'type': 'http.disconnect'
        })


class RequestHandler:
    """A request handler"""

    def __init__(
            self,
            url: urllib.parse.ParseResult,
            method: str,
            headers: Optional[List[Header]],
            content: Optional[AsyncIterable[bytes]],
            decompressors: Decompressors
    ) -> None:
        """Initialise the request handler

        Args:
            url (urllib.parse.ParseResult): The parsed url
            method (str): The request method
            headers (Optional[List[Header]]): The headers
            content (Optional[AsyncIterable[bytes]]): The request body
            decompressors (Decompressors]): The decompressors
        """
        self.url = url
        self.method = method
        self.headers = headers or []
        self.content = content
        self.instance: Optional[RequestHandlerInstance] = None
        self.decompressors = decompressors

    async def __call__(
            self,
            receive: ReceiveCallable,
            send: SendCallable
    ) -> Mapping[str, Any]:
        """Call the request handle instance

        Args:
            receive (ReceiveCallable): The function to receive data
            send (SendCallable): The function to send data

        Returns:
            Mapping[str, Any]: [description]
        """
        self.instance = RequestHandlerInstance(
            self.url,
            self.method,
            self.headers,
            self.content,
            send,
            receive,
            self.decompressors
        )
        response = await self.instance.process()
        return response

    async def close(self):
        """Close the request"""
        if self.instance is not None:
            await self.instance.close()



"""A handler for the HTTP/1.1 protocol"""

import asyncio
from typing import (
    Any,
    Mapping,
    Optional
)

import h11

from .http_protocol import HttpProtocol
from .utils import get_target
from .asyncio_events import MessageEvent

MappingMessageEvent = MessageEvent[Mapping[str, Any]]


class H11Protocol(HttpProtocol):
    """An HTTP/1.1 protocol handler"""

    def __init__(
            self,
            reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter,
            bufsiz: int
    ) -> None:
        """Initialise an HTTP/1.1 protocol handler

        Args:
            reader (asyncio.StreamReader): The data reader
            writer (asyncio.StreamWriter): The data writer
            bufsiz (int): The buffer size used when receiving data.
        """
        super().__init__(reader, writer)
        self._bufsiz = bufsiz
        self._h11_state = h11.Connection(our_role=h11.CLIENT)
        self._is_initialised = False
        self._connection_event: MappingMessageEvent = MessageEvent()
        self._response_event: MappingMessageEvent = MessageEvent()
        self._is_message_ended = True

    def _connect(self) -> None:
        if self._is_initialised:
            self._h11_state.start_next_cycle()
        self._is_message_ended = False

    async def send(self, message: Mapping[str, Any]) -> None:

        request_type: str = message['type']

        if request_type == 'http.request':
            await self._send_request(message)
        elif request_type == 'http.request.body':
            await self._send_request_body(message)
        elif request_type == 'http.disconnect':
            await self._disconnect()
        else:
            raise Exception(f'unknown request type: {request_type}')

    async def receive(self) -> Mapping[str, Any]:

        message = await self._connection_event.wait_with_message()
        if message is not None:
            return message

        message = await self._response_event.wait_with_message()
        if message is not None:
            return message

        message = await self._receive_body_event()
        return message

    async def _send_request(self, message: Mapping[str, Any]) -> None:

        self._connect()
        self._is_initialised = True

        request = h11.Request(
            method=message['method'],
            target=get_target(message['url']),
            headers=message.get('headers', [])
        )

        buf = self._h11_state.send(request)
        self.writer.write(buf)
        await self.writer.drain()
        self._connection_event.set_with_message({
            'type': 'http.response.connection',
            'http_version': 'h11',
            'stream_id': None
        })

        body = message.get('body', b'')
        more_body = message.get('more_body', False)
        await self._send_request_data(body, more_body)
        asyncio.create_task(self._receive_response())

    async def _send_request_body(self, message: Mapping[str, Any]) -> None:
        await self._send_request_data(
            message.get('body', b''),
            message.get('more_body', False)
        )

    async def _send_request_data(
            self,
            body: bytes,
            more_body: Optional[bool]
    ) -> None:
        buf = self._h11_state.send(h11.Data(data=body))
        self.writer.write(buf)

        if not more_body:
            buf = self._h11_state.send(h11.EndOfMessage())
            self.writer.write(buf)

        await self.writer.drain()

    async def _receive_response(self) -> None:
        while True:
            event = self._h11_state.next_event()
            if event is h11.NEED_DATA:
                buf = await self.reader.read(self._bufsiz)
                self._h11_state.receive_data(buf)
            elif isinstance(event, h11.Response):
                break
            elif isinstance(event, (h11.ConnectionClosed, h11.EndOfMessage)):
                raise ConnectionError('Failed to receive response')
            else:
                raise ValueError('Unknown event')

        more_body = False
        for name, value in event.headers:
            if name == b'content-length' and int(value):
                more_body = True
            elif name == b'transfer-encoding' and value == b'chunked':
                more_body = True

        self._response_event.set_with_message({
            'type': 'http.response',
            'acgi': {
                'version': "1.0"
            },
            'http_version': '1.1',
            'status_code': event.status_code,
            'headers': event.headers,
            'more_body': more_body,
            'stream_id': None
        })

    async def _disconnect(self) -> None:
        if not self._is_message_ended and self._h11_state.our_state != h11.DONE:

            if self._h11_state.our_state == h11.MUST_CLOSE:
                buf = self._h11_state.send(h11.ConnectionClosed())
            else:
                buf = self._h11_state.send(h11.EndOfMessage())

            if buf:
                self.writer.write(buf)
                await self.writer.drain()

        self.writer.close()
        await self.writer.wait_closed()

    async def _receive_body_event(self) -> Mapping[str, Any]:
        while True:
            event = self._h11_state.next_event()
            if event is h11.NEED_DATA:
                self._h11_state.receive_data(await self.reader.read(self._bufsiz))
            elif isinstance(event, h11.Data):
                return {
                    'type': 'http.response.body',
                    'body': event.data,
                    'more_body': True,
                    'stream_id': None
                }
            elif isinstance(event, h11.EndOfMessage):
                self._is_message_ended = True
                return {
                    'type': 'http.response.body',
                    'body': b'',
                    'more_body': False,
                    'stream_id': None
                }
            elif isinstance(event, h11.ConnectionClosed):
                return {
                    'type': 'http.stream.disconnect',
                    'stream_id': None
                }
            else:
                raise ValueError('Unknown event')



"""Helpers"""

from asyncio import AbstractEventLoop
import json
import ssl
from typing import Any, Callable, Iterable, Mapping, Optional, Type, Union
from urllib.error import HTTPError
from urllib.parse import urlparse

from baretypes import Headers
import bareutils.header as header
from bareutils import bytes_writer
from bareutils.compression import Decompressor

from .client import HttpClient
from .constants import USER_AGENT, DEFAULT_PROTOCOLS
from .ssl_contexts import DEFAULT_CIPHERS, DEFAULT_OPTIONS


async def request(
        url: str,
        method: str,
        *,
        headers: Headers = None,
        content: Optional[bytes] = None,
        loop: Optional[AbstractEventLoop] = None,
        cafile: Optional[str] = None,
        capath: Optional[str] = None,
        cadata: Optional[str] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        decompressors: Optional[Mapping[bytes, Type[Decompressor]]] = None,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS,
        chunk_size: int = -1,
        connect_timeout: Optional[Union[int, float]] = None
) -> bytes:
    """Gets bytes from a url.

    ```python
    buf = await request(
        'https://jsonplaceholder.typicode.com/todos/1',
        'GET',
        ssl=ssl.SSLContext()
    )
    ```

    Args:
        url (str): The url to get.
        method (str): The HTTP method (eg. 'GET', 'POST', etc).
        headers (Headers, optional): Any extra headers required. Defaults to
            None.
        content (Optional[bytes], optional): The content to send.. Defaults to
            None.
        loop (Optional[AbstractEventLoop], optional): The optional asyncio event
            loop.. Defaults to None.
        cafile (Optional[str], optional): The path to a file of concatenated CA
            certificates in PEM format. Defaults to None.
        capath (Optional[str], optional): The path to a directory containing
            several CA certificates in PEM format. Defaults to None.
        cadata (Optional[str], optional): Either an ASCII string of one or more
            PEM-encoded certificates or a bytes-like object of DER-encoded
            certificates. Defaults to None.
        ssl_context (Optional[SSLContext], optional): An ssl context to be
            used instead of generating one from the certificates.
        decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
            The decompressors. Defaults to None.
        protocols (Iterable[str], optional): The supported protocols. Defaults
            to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The supported ciphers. Defaults
            to DEFAULT_CIPHERS.
        options (Iterable[int], optional): The ssl.SSLContext.options. Defaults
            to DEFAULT_OPTIONS.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk. Defaults to -1.
        connect_timeout (Optional[Union[int, float]], optional): The number
            of seconds to wait for the connection. Defaults to None.

    Raises:
        HTTPError: Is the status code is not ok.
        asyncio.TimeoutError: If the connect times out.

    Returns:
        bytes: The bytes received.
    """

    headers = [] if headers is None else list(headers)

    content_length = str(len(content)).encode('ascii') if content else b'0'

    hostname = urlparse(url).hostname
    if hostname is None:
        raise RuntimeError('unspecified hostname')

    base_headers = [
        (b'host', hostname.encode('ascii')),
        (b'content-length', content_length),
        (b'user-agent', USER_AGENT),
        (b'connection', b'close')
    ]

    for name, value in base_headers:
        if not header.find(name, headers):
            headers.append((name, value))

    data = bytes_writer(content, chunk_size) if content else None

    async with HttpClient(
            url,
            method=method,
            headers=headers,
            content=data,
            loop=loop,
            cafile=cafile,
            capath=capath,
            cadata=cadata,
            ssl_context=ssl_context,
            decompressors=decompressors,
            protocols=protocols,
            ciphers=ciphers,
            options=options,
            connect_timeout=connect_timeout
    ) as response:
        buf = b''
        if response['more_body']:
            async for part in response['body']:
                buf += part
        if response['status_code'] < 200 or response['status_code'] >= 400:
            raise HTTPError(
                url,
                response['status_code'],
                buf.decode(),
                response['headers'],
                None
            )
        return buf


async def get(
        url: str,
        *,
        headers: Headers = None,
        loop: Optional[AbstractEventLoop] = None,
        cafile: Optional[str] = None,
        capath: Optional[str] = None,
        cadata: Optional[str] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        decompressors: Optional[Mapping[bytes, Type[Decompressor]]] = None,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS,
        connect_timeout: Optional[Union[int, float]] = None
) -> bytes:
    """Issues a GET request

    Args:
        url (str): The url
        headers (Headers, optional): Any extra headers required. Defaults to
            None.
        loop (Optional[AbstractEventLoop], optional): The optional asyncio event
            loop.. Defaults to None.
        cafile (Optional[str], optional): The path to a file of concatenated CA
            certificates in PEM format. Defaults to None.
        capath (Optional[str], optional): The path to a directory containing
            several CA certificates in PEM format. Defaults to None.
        cadata (Optional[str], optional): Either an ASCII string of one or more
            PEM-encoded certificates or a bytes-like object of DER-encoded
            certificates. Defaults to None.
        ssl_context (Optional[SSLContext], optional): An ssl context to be
            used instead of generating one from the certificates.
        decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
            The decompressors. Defaults to None.
        protocols (Iterable[str], optional): The supported protocols. Defaults
            to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The supported ciphers. Defaults
            to DEFAULT_CIPHERS.
        options (Iterable[int], optional): The ssl.SSLContext.options. Defaults
            to DEFAULT_OPTIONS.
        connect_timeout (Optional[Union[int, float]], optional): The number
            of seconds to wait for the connection. Defaults to None.

    Raises:
        HTTPError: Is the status code is not ok.
        asyncio.TimeoutError: If the connect times out.

    Returns:
        bytes: [description]
    """
    return await request(
        url,
        'GET',
        headers=headers,
        loop=loop,
        cafile=cafile,
        capath=capath,
        cadata=cadata,
        ssl_context=ssl_context,
        decompressors=decompressors,
        protocols=protocols,
        ciphers=ciphers,
        options=options,
        connect_timeout=connect_timeout
    )


async def get_text(
        url: str,
        *,
        headers: Headers = None,
        loop: Optional[AbstractEventLoop] = None,
        encoding: str = 'utf-8',
        cafile: Optional[str] = None,
        capath: Optional[str] = None,
        cadata: Optional[str] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        decompressors: Optional[Mapping[bytes, Type[Decompressor]]] = None,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS,
        connect_timeout: Optional[Union[int, float]] = None
) -> str:
    """Issues a GET request returning a string

    The following gets some text:

    ```python
    import asyncio
    from bareclient import get_text

    async def main(url: str) -> None:
        text = await get_text(url)
        print(text)

    asyncio.run(main('https://docs.python.org/3/library/cgi.html'))
    ```

    Args:
        url (str): The url
        headers (Headers, optional): Any extra headers required. Defaults to
            None.
        loop (Optional[AbstractEventLoop], optional): The optional asyncio event
            loop.. Defaults to None.
        cafile (Optional[str], optional): The path to a file of concatenated CA
            certificates in PEM format. Defaults to None.
        capath (Optional[str], optional): The path to a directory containing
            several CA certificates in PEM format. Defaults to None.
        cadata (Optional[str], optional): Either an ASCII string of one or more
            PEM-encoded certificates or a bytes-like object of DER-encoded
            certificates. Defaults to None.
        ssl_context (Optional[SSLContext], optional): An ssl context to be
            used instead of generating one from the certificates.
        decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
            The decompressors. Defaults to None.
        protocols (Iterable[str], optional): The supported protocols. Defaults
            to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The supported ciphers. Defaults
            to DEFAULT_CIPHERS.
        options (Iterable[int], optional): The ssl.SSLContext.options. Defaults
            to DEFAULT_OPTIONS.
        connect_timeout (Optional[Union[int, float]], optional): The number
            of seconds to wait for the connection. Defaults to None.

    Raises:
        HTTPError: Is the status code is not ok.
        asyncio.TimeoutError: If the connect times out.

    Returns:
        str: [description]
    """

    headers = [] if headers is None else list(headers)

    if not header.find(b'accept', headers):
        headers.append((b'accept', b'text/plain'))

    buf = await get(
        url,
        headers=headers,
        loop=loop,
        cafile=cafile,
        capath=capath,
        cadata=cadata,
        ssl_context=ssl_context,
        decompressors=decompressors,
        protocols=protocols,
        ciphers=ciphers,
        options=options,
        connect_timeout=connect_timeout
    )
    return buf.decode(encoding)


async def get_json(
        url: str,
        *,
        headers: Headers = None,
        loads: Callable[[str], Any] = json.loads,
        loop: Optional[AbstractEventLoop] = None,
        cafile: Optional[str] = None,
        capath: Optional[str] = None,
        cadata: Optional[str] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        decompressors: Optional[Mapping[bytes, Type[Decompressor]]] = None,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS,
        connect_timeout: Optional[Union[int, float]] = None
) -> Any:
    """Issues a GET request returning a JSON object

    The following gets some json:

    ```python
    import asyncio
    from bareclient import get_json

    async def main(url: str) -> None:
        obj = await get_json(url)
        print(obj)

    asyncio.run(main('https://jsonplaceholder.typicode.com/todos/1'))
    ```

    Args:
        url (str): The url
        headers (Headers, optional): Any extra headers required. Defaults to
            None.
        loads (Callable[[str], Any], optional): The function to loads the
            JSON object from the string. Defaults to json.loads.
        loop (Optional[AbstractEventLoop], optional): The optional asyncio event
            loop.. Defaults to None.
        cafile (Optional[str], optional): The path to a file of concatenated CA
            certificates in PEM format. Defaults to None.
        capath (Optional[str], optional): The path to a directory containing
            several CA certificates in PEM format. Defaults to None.
        cadata (Optional[str], optional): Either an ASCII string of one or more
            PEM-encoded certificates or a bytes-like object of DER-encoded
            certificates. Defaults to None.
        ssl_context (Optional[SSLContext], optional): An ssl context to be
            used instead of generating one from the certificates.
        decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
            The decompressors. Defaults to None.
        protocols (Iterable[str], optional): The supported protocols. Defaults
            to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The supported ciphers. Defaults
            to DEFAULT_CIPHERS.
        options (Iterable[int], optional): The ssl.SSLContext.options. Defaults
            to DEFAULT_OPTIONS.
        connect_timeout (Optional[Union[int, float]], optional): The number
            of seconds to wait for the connection. Defaults to None.

    Raises:
        HTTPError: Is the status code is not ok.
        asyncio.TimeoutError: If the connect times out.

    Returns:
        Any: The decoded JSON object
    """
    headers = [] if headers is None else list(headers)

    if not header.find(b'accept', headers):
        headers.append((b'accept', b'application/json'))

    text = await get_text(
        url,
        headers=headers,
        encoding='utf-8',
        loop=loop,
        cafile=cafile,
        capath=capath,
        cadata=cadata,
        ssl_context=ssl_context,
        decompressors=decompressors,
        protocols=protocols,
        ciphers=ciphers,
        options=options,
        connect_timeout=connect_timeout
    )
    return loads(text)


async def post(
        url: str,
        content: bytes,
        *,
        headers: Headers = None,
        loop: Optional[AbstractEventLoop] = None,
        cafile: Optional[str] = None,
        capath: Optional[str] = None,
        cadata: Optional[str] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        decompressors: Optional[Mapping[bytes, Type[Decompressor]]] = None,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS,
        chunk_size: int = -1,
        connect_timeout: Optional[Union[int, float]] = None
) -> bytes:
    """Issues a POST request

    Args:
        url (str): The url
        content (bytes): The body content
        headers (Headers, optional): Any extra headers required. Defaults to
            None.
        loop (Optional[AbstractEventLoop], optional): The optional asyncio event
            loop.. Defaults to None.
        cafile (Optional[str], optional): The path to a file of concatenated CA
            certificates in PEM format. Defaults to None.
        capath (Optional[str], optional): The path to a directory containing
            several CA certificates in PEM format. Defaults to None.
        cadata (Optional[str], optional): Either an ASCII string of one or more
            PEM-encoded certificates or a bytes-like object of DER-encoded
            certificates. Defaults to None.
        ssl_context (Optional[SSLContext], optional): An ssl context to be
            used instead of generating one from the certificates.
        decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
            The decompressors. Defaults to None.
        protocols (Iterable[str], optional): The supported protocols. Defaults
            to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The supported ciphers. Defaults
            to DEFAULT_CIPHERS.
        options (Iterable[int], optional): The ssl.SSLContext.options. Defaults
            to DEFAULT_OPTIONS.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk.. Defaults to -1.
        connect_timeout (Optional[Union[int, float]], optional): The number
            of seconds to wait for the connection. Defaults to None.

    Raises:
        HTTPError: Is the status code is not ok.
        asyncio.TimeoutError: If the connect times out.

    Returns:
        bytes: The response body
    """
    return await request(
        url,
        method='POST',
        content=content,
        headers=headers,
        loop=loop,
        cafile=cafile,
        capath=capath,
        cadata=cadata,
        ssl_context=ssl_context,
        decompressors=decompressors,
        protocols=protocols,
        ciphers=ciphers,
        options=options,
        chunk_size=chunk_size,
        connect_timeout=connect_timeout
    )


async def post_text(
        url: str,
        text: str,
        *,
        encoding='utf-8',
        headers: Headers = None,
        loop: Optional[AbstractEventLoop] = None,
        cafile: Optional[str] = None,
        capath: Optional[str] = None,
        cadata: Optional[str] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        decompressors: Optional[Mapping[bytes, Type[Decompressor]]] = None,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS,
        chunk_size: int = -1,
        connect_timeout: Optional[Union[int, float]] = None
) -> str:
    """Issues a POST request with a str body

    Args:
        url (str): The url
        content (bytes): The body content
        headers (Headers, optional): Any extra headers required. Defaults to
            None.
        loop (Optional[AbstractEventLoop], optional): The optional asyncio event
            loop.. Defaults to None.
        cafile (Optional[str], optional): The path to a file of concatenated CA
            certificates in PEM format. Defaults to None.
        capath (Optional[str], optional): The path to a directory containing
            several CA certificates in PEM format. Defaults to None.
        cadata (Optional[str], optional): Either an ASCII string of one or more
            PEM-encoded certificates or a bytes-like object of DER-encoded
            certificates. Defaults to None.
        ssl_context (Optional[SSLContext], optional): An ssl context to be
            used instead of generating one from the certificates.
        decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
            The decompressors. Defaults to None.
        protocols (Iterable[str], optional): The supported protocols. Defaults
            to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The supported ciphers. Defaults
            to DEFAULT_CIPHERS.
        options (Iterable[int], optional): The ssl.SSLContext.options. Defaults
            to DEFAULT_OPTIONS.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk.. Defaults to -1.
        connect_timeout (Optional[Union[int, float]], optional): The number
            of seconds to wait for the connection. Defaults to None.

    Raises:
        HTTPError: Is the status code is not ok.
        asyncio.TimeoutError: If the connect times out.

    Returns:
        bytes: The response body
    """

    headers = [] if headers is None else list(headers)

    if not header.find(b'accept', headers):
        headers.append((b'accept', b'text/plain'))

    content = text.encode(encoding=encoding)
    if not header.find(b'content-type', headers):
        headers.append((b'content-type', b'text/plain'))

    response = await post(
        url,
        content=content,
        headers=headers,
        loop=loop,
        cafile=cafile,
        capath=capath,
        cadata=cadata,
        ssl_context=ssl_context,
        decompressors=decompressors,
        protocols=protocols,
        ciphers=ciphers,
        options=options,
        chunk_size=chunk_size,
        connect_timeout=connect_timeout
    )
    return response.decode(encoding=encoding)


async def post_json(
        url: str,
        obj: Any,
        *,
        loads: Callable[[str], Any] = json.loads,
        dumps: Callable[[Any], str] = json.dumps,
        headers: Headers = None,
        loop: Optional[AbstractEventLoop] = None,
        cafile: Optional[str] = None,
        capath: Optional[str] = None,
        cadata: Optional[str] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        decompressors: Optional[Mapping[bytes, Type[Decompressor]]] = None,
        protocols: Iterable[str] = DEFAULT_PROTOCOLS,
        ciphers: Iterable[str] = DEFAULT_CIPHERS,
        options: Iterable[int] = DEFAULT_OPTIONS,
        chunk_size: int = -1,
        connect_timeout: Optional[Union[int, float]] = None
) -> Optional[Any]:
    """Issues a POST request with a JSON payload

    ```python
    import asyncio
    from bareclient import post_json

    async def main(url: str) -> None:
        obj = await post_json(
            url,
            {'title': 'A job'},
            headers=[(b'accept-encoding', b'gzip')]
        )
        print(obj)

    asyncio.run(main('https://jsonplaceholder.typicode.com/todos'))
    ```

    Args:
        url (str): The url
        obj (Any): The JSON payload
        loads (Callable[[str], Any], optional): The function used to decode
            the response. Defaults to json.loads.
        dumps (Callable[[Any], str], optional): The function used to encode
            the request. Defaults to json.dumps.
        headers (Headers, optional): Any extra headers required. Defaults to
            None.
        loop (Optional[AbstractEventLoop], optional): The optional asyncio event
            loop.. Defaults to None.
        cafile (Optional[str], optional): The path to a file of concatenated CA
            certificates in PEM format. Defaults to None.
        capath (Optional[str], optional): The path to a directory containing
            several CA certificates in PEM format. Defaults to None.
        cadata (Optional[str], optional): Either an ASCII string of one or more
            PEM-encoded certificates or a bytes-like object of DER-encoded
            certificates. Defaults to None.
        ssl_context (Optional[SSLContext], optional): An ssl context to be
            used instead of generating one from the certificates.
        decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
            The decompressors. Defaults to None.
        protocols (Iterable[str], optional): The supported protocols. Defaults
            to DEFAULT_PROTOCOLS.
        ciphers (Iterable[str], optional): The supported ciphers. Defaults
            to DEFAULT_CIPHERS.
        options (Iterable[int], optional): The ssl.SSLContext.options. Defaults
            to DEFAULT_OPTIONS.
        chunk_size (int, optional): The size of each chunk to send or -1 to send
            as a single chunk.. Defaults to -1.
        connect_timeout (Optional[Union[int, float]], optional): The number
            of seconds to wait for the connection. Defaults to None.

    Raises:
        HTTPError: Is the status code is not ok.
        asyncio.TimeoutError: If the connect times out.

    Returns:
        Optional[Any]: The decoded response
    """

    headers = [] if headers is None else list(headers)

    if not header.find(b'accept', headers):
        headers.append((b'accept', b'application/json'))

    content = dumps(obj)
    if not header.find(b'content-type', headers):
        headers.append((b'content-type', b'application/json'))

    text = await post_text(
        url,
        text=content,
        encoding='utf-8',
        headers=headers,
        loop=loop,
        cafile=cafile,
        capath=capath,
        cadata=cadata,
        ssl_context=ssl_context,
        decompressors=decompressors,
        protocols=protocols,
        ciphers=ciphers,
        options=options,
        chunk_size=chunk_size,
        connect_timeout=connect_timeout
    )
    return loads(text)



"""The HTTP Client"""

from asyncio import AbstractEventLoop
import urllib.parse
from ssl import SSLContext
from typing import (
    Any,
    Iterable,
    List,
    Mapping,
    Optional,
    Union
)

from baretypes import Header, Content

from .requester import RequestHandler
from .acgi import connect
from .constants import DEFAULT_DECOMPRESSORS, DEFAULT_PROTOCOLS, Decompressors
from .ssl_contexts import DEFAULT_CIPHERS, DEFAULT_OPTIONS


class HttpClient:
    """An HTTP client"""

    def __init__(
            self,
            url: str,
            *,
            method: str = 'GET',
            headers: Optional[List[Header]] = None,
            content: Optional[Content] = None,
            loop: Optional[AbstractEventLoop] = None,
            h11_bufsiz: int = 8096,
            cafile: Optional[str] = None,
            capath: Optional[str] = None,
            cadata: Optional[str] = None,
            ssl_context: Optional[SSLContext] = None,
            decompressors: Optional[Decompressors] = None,
            protocols: Iterable[str] = DEFAULT_PROTOCOLS,
            ciphers: Iterable[str] = DEFAULT_CIPHERS,
            options: Iterable[int] = DEFAULT_OPTIONS,
            connect_timeout: Optional[Union[int, float]] = None
    ) -> None:
        """Make an HTTP client.

        The following example will make a GET request.

        ```python
        import asyncio
        from bareclient import HttpClient


        async def main(url: str) -> None:
            async with HttpClient(url, method='GET') as response:
                print(response)
                if response['status_code'] == 200 and response['more_body']:
                    async for part in response['body']:
                        print(part)

        asyncio.run(main('https://docs.python.org/3/library/cgi.html'))
        ```

        The `response` is a mapping with the following content:

        - **`type`** (_Unicode string_) - Currently the only response is
        `"http.response"`.
        - **`acgi["version"]`** (_Unicode string_) - Version of the ACGI spec.
        - **`http_version`** (_Unicode string_) - One of `"1.0"`, `"1.1"` or `"2"`.
        - **`stream_id`** (_int_) - The HTTP/2 stream id, otherwise None.
        - **`status_code`** (_int_) - The HTTP status code.
        - **`headers`** (_Iterable[[byte string, byte string]]_) - A iterable of [name,
        value] two-item iterables, where name is the header name, and value is the
        header value. Order must be preserved in the HTTP response. Header names
        must be lowercased. Optional; defaults to an empty list. Pseudo headers
        (present in HTTP/2 and HTTP/3) must not be present.
        - **`more_body`** (_bool_) - Signifies if the body has more content.
        - **`body`** (_AsyncIterable[byte string]_) - The body content.

        Args:
            url (str): The url
            method (str, optional): The HTTP method. Defaults to 'GET'.
            headers (Optional[List[Header]], optional): The headers. Defaults to
                None.
            content (Optional[Content], optional): The body content. Defaults to
                None.
            loop (Optional[AbstractEventLoop], optional): The optional asyncio
                event loop. Defaults to None.
            h11_bufsiz (int, optional): The HTTP/1 buffer size. Defaults to
                8096.
            cafile (Optional[str], optional): The path to a file of concatenated
                CA certificates in PEM format. Defaults to None.
            capath (Optional[str], optional): The path to a directory containing
                several CA certificates in PEM format. Defaults to None.
            cadata (Optional[str], optional): Either an ASCII string of one or
                more PEM-encoded certificates or a bytes-like object of
                DER-encoded certificates. Defaults to None.
            ssl_context (Optional[SSLContext], optional): An ssl context to be
                used instead of generating one from the certificates.
            decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
                The decompressors. Defaults to None.
            protocols (Iterable[str], optional): The supported protocols. Defaults
                to DEFAULT_PROTOCOLS.
            ciphers (Iterable[str], optional): The supported ciphers. Defaults
                to DEFAULT_CIPHERS.
            options (Iterable[int], optional): The ssl.SSLContext.options. Defaults
                to DEFAULT_OPTIONS.
            connect_timeout (Optional[Union[int, float]], optional): The number
                of seconds to wait for the connection. Defaults to None.
        """
        self.url = urllib.parse.urlparse(url)
        self.method = method
        self.headers = headers
        self.content = content
        self.loop = loop
        self.h11_bufsiz = h11_bufsiz
        self.cafile = cafile
        self.capath = capath
        self.cadata = cadata
        self.ssl_context = ssl_context
        self.handler: Optional[RequestHandler] = None
        self.decompressors = decompressors or DEFAULT_DECOMPRESSORS
        self.protocols = protocols
        self.ciphers = ciphers
        self.options = options
        self.connect_timeout = connect_timeout

    async def __aenter__(self) -> Mapping[str, Any]:
        self.handler = RequestHandler(
            self.url,
            self.method,
            self.headers,
            self.content,
            self.decompressors
        )
        response = await connect(
            self.url,
            self.handler,
            cafile=self.cafile,
            capath=self.capath,
            cadata=self.cadata,
            ssl_context=self.ssl_context,
            loop=self.loop,
            h11_bufsiz=self.h11_bufsiz,
            protocols=self.protocols,
            ciphers=self.ciphers,
            options=self.options,
            connect_timeout=self.connect_timeout
        )
        return response

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.handler is not None:
            await self.handler.close()



"""A handler for HTTP protocols"""

from abc import ABCMeta, abstractmethod
from asyncio import StreamReader, StreamWriter
from typing import Any, Mapping


class HttpProtocol(metaclass=ABCMeta):
    """The base class for HTTP protocol handlers"""

    def __init__(
            self,
            reader: StreamReader,
            writer: StreamWriter
    ) -> None:
        """Initialise an HTTP protocol

        Args:
            reader (StreamReader): The reader
            writer (StreamWriter): The writer
        """
        self.reader = reader
        self.writer = writer

    @abstractmethod
    async def send(self, message: Mapping[str, Any]) -> None:
        """Send a message to the web server

        Args:
            message (Mapping[str, Any]): The message to send
        """

    @abstractmethod
    async def receive(self) -> Mapping[str, Any]:
        """Receive a message from the web server

        Returns:
            Mapping[str, Any]: The message received
        """



"""A handler for the HTTP/2 protocol"""

import asyncio
from asyncio import Task
import functools
from typing import (
    Any,
    Awaitable,
    Callable,
    List,
    Mapping,
    MutableMapping,
    Optional
)
import urllib.parse

import h2.connection
import h2.events
import h2.settings

from baretypes import Header

from .http_protocol import HttpProtocol
from .utils import get_target, get_authority
from .asyncio_events import ResetEvent


class H2Protocol(HttpProtocol):
    """An HTTP/2 protocol handler"""

    READ_NUM_BYTES = 4096

    def __init__(
            self,
            reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter
    ) -> None:
        """Initialise an HTTP/2 protocol

        Args:
            reader (asyncio.StreamReader): The reader
            writer (asyncio.StreamWriter): The writer
        """
        super().__init__(reader, writer)
        self.h2_state = h2.connection.H2Connection()
        self.window_update_event: MutableMapping[int, ResetEvent] = {}
        self.initialized = False
        self.responses: asyncio.Queue = asyncio.Queue()
        self.response_task: Optional[asyncio.Task] = None
        self.pending: List[Task] = []
        self.on_close: Optional[Callable[[], Awaitable[None]]] = None
        self.h2_events: List[h2.events.Event] = []

    async def send(
            self,
            message: Mapping[str, Any]
    ) -> None:
        message_type: str = message['type']

        if message_type == 'http.request':
            await self._send_request(message)
        elif message_type == 'http.request.body':
            await self._send_request_body(message)
        elif message_type == 'http.disconnect':
            if self.on_close:
                await self.on_close()
        else:
            raise RuntimeError(f'unknown request type: {message_type}')

    async def receive(self) -> Mapping[str, Any]:
        return await self.responses.get()

    async def _send_request(
            self,
            message: Mapping[str, Any]
    ) -> None:
        # Start sending the request.
        if not self.initialized:
            self._initiate_connection()

        stream_id = await self._send_headers(
            message['url'],
            message['method'],
            message.get('headers', [])
        )
        self.window_update_event[stream_id] = ResetEvent()
        await self.responses.put({
            'type': 'http.response.connection',
            'http_version': 'h2',
            'stream_id': stream_id
        })

        body = message.get('body', b'')
        more_body = message.get('more_body', False)

        self._create_task(
            self._send_request_data(stream_id, body, more_body)
        )
        self.response_task = asyncio.create_task(self._receive_response())
        self.on_close = functools.partial(
            self._response_closed, stream_id=stream_id)

    async def _send_request_body(
            self,
            message: Mapping[str, Any]
    ) -> None:
        await self._send_request_data(
            message['stream_id'],
            message['body'],
            message.get('more_body', False)
        )

    def _create_task(self, coroutine) -> Task:
        task = asyncio.create_task(coroutine)
        self.pending.append(task)
        task.add_done_callback(self._reap_task)
        return task

    def _reap_task(self, task: Task):
        self.pending.remove(task)

    def _initiate_connection(self) -> None:
        self.h2_state.local_settings = h2.settings.Settings(
            client=True,
            initial_values={
                h2.settings.SettingCodes.ENABLE_PUSH: 0,
                h2.settings.SettingCodes.MAX_CONCURRENT_STREAMS: 100,
                h2.settings.SettingCodes.MAX_HEADER_LIST_SIZE: 65536,
            },
        )

        del self.h2_state.local_settings[
            h2.settings.SettingCodes.ENABLE_CONNECT_PROTOCOL
        ]

        self.h2_state.initiate_connection()
        data_to_send = self.h2_state.data_to_send()
        self.writer.write(data_to_send)
        self.initialized = True

    async def _send_headers(
            self,
            url: urllib.parse.ParseResult,
            method: str,
            headers: List[Header]
    ) -> int:
        stream_id = self.h2_state.get_next_available_stream_id()
        headers = [
            (b":method", method.encode("ascii")),
            (b":authority", get_authority(url).encode("ascii")),
            (b":scheme", url.scheme.encode("ascii")),
            (b":path", get_target(url).encode("ascii")),
        ] + headers

        self.h2_state.send_headers(stream_id, headers)
        data_to_send = self.h2_state.data_to_send()
        self.writer.write(data_to_send)
        await self.writer.drain()

        return stream_id

    async def _send_request_data(
            self,
            stream_id: int,
            body: bytes,
            more_body: bool
    ) -> None:
        await self._send_data(stream_id, body)
        if not more_body:
            await self._end_stream(stream_id)

    async def _send_data(
            self,
            stream_id: int,
            data: bytes
    ) -> None:
        while data:
            window_size = self.h2_state.local_flow_control_window(stream_id)
            chunk_size = min(
                len(data),
                window_size,
                self.h2_state.max_outbound_frame_size
            )
            if chunk_size == 0:
                await self.window_update_event[stream_id].wait()
            else:
                chunk, data = data[:chunk_size], data[chunk_size:]
                self.h2_state.send_data(stream_id, chunk)
                data_to_send = self.h2_state.data_to_send()
                self.writer.write(data_to_send)
                await self.writer.drain()

    async def _end_stream(self, stream_id: int) -> None:
        self.h2_state.end_stream(stream_id)
        data_to_send = self.h2_state.data_to_send()
        self.writer.write(data_to_send)
        await self.writer.drain()

    async def _receive_response(self) -> None:

        event: Optional[h2.events.Event] = None
        while not isinstance(event, h2.events.ResponseReceived):
            event = await self._receive_event()

        status_code = 200
        headers: List[Header] = []
        for name, value in event.headers:
            if name == b":status":
                status_code = int(value)
            elif not name.startswith(b":"):
                headers.append((name, value))

        await self.responses.put({
            'type': 'http.response',
            'acgi': {
                'version': "1.0"
            },
            'http_version': '2',
            'status_code': status_code,
            'headers': headers,
            'more_body': event.stream_ended is None,
            'stream_id': event.stream_id
        })

        is_connected = True
        while is_connected:
            event = await self._receive_event()
            if isinstance(event, h2.events.DataReceived):
                self.h2_state.acknowledge_received_data(
                    event.flow_controlled_length, event.stream_id
                )
                await self.responses.put({
                    'type': 'http.response.body',
                    'body': event.data,
                    'more_body': event.stream_ended is None,
                    'stream_id': event.stream_id
                })
            elif isinstance(event, (h2.events.StreamEnded, h2.events.StreamReset)):
                await self.responses.put({
                    'type': 'http.stream.disconnect',
                    'stream_id': event.stream_id
                })
                is_connected = False

    async def _receive_event(self) -> h2.events.Event:
        while not self.h2_events:
            data = await self.reader.read(self.READ_NUM_BYTES)
            events = self.h2_state.receive_data(data)
            for event in events:
                if isinstance(event, h2.events.WindowUpdated):
                    if event.stream_id == 0:
                        for update_event in self.window_update_event.values():
                            update_event.set()
                    else:
                        self.window_update_event[event.stream_id].set()

                self.h2_events.append(event)

            data_to_send = self.h2_state.data_to_send()
            self.writer.write(data_to_send)
            await self.writer.drain()

        return self.h2_events.pop(0)

    async def _response_closed(self, stream_id: int) -> None:
        del self.window_update_event[stream_id]

        # Drain responses to allow the socket to close cleanly.
        if self.response_task is not None:
            await self.response_task
            while self.responses.qsize():
                await self.responses.get()

        for task in self.pending:
            if not task.done():
                try:
                    task.cancel()
                    await task
                except asyncio.CancelledError:
                    pass

        self.writer.close()
        await self.writer.wait_closed()



"""Utilities"""

from asyncio import StreamWriter
import logging
import ssl
from typing import Optional
from urllib.parse import ParseResult

LOGGER = logging.getLogger(__name__)

SCHEMES = {
    'http': {
        'port': 80
    },
    'https': {
        'port': 443
    }
}


def get_port(url: ParseResult) -> Optional[int]:
    """Gets the port.

    Args:
        url (ParseResult): A parsed url.

    Raises:
        ValueError: If the scheme is unknown.

    Returns:
        Optional[int]: The port.
    """
    if url.scheme not in SCHEMES:
        raise ValueError('unknown scheme')
    return url.port if url.port else SCHEMES[url.scheme]['port']


def get_target(url: ParseResult) -> str:
    """Gets the target.

    Args:
        url (ParseResult): A parsed url

    Returns:
        str: The target.
    """
    path = url.path
    if url.query:
        path += '?' + url.query
    if url.fragment:
        path += '#' + url.fragment
    return path


def get_authority(url: ParseResult) -> str:
    """Get the http/2 authority"""
    if isinstance(url.netloc, str):
        return url.netloc
    host, _port = url.netloc.split(':', maxsplit=1)
    return host


def get_negotiated_protocol(writer: StreamWriter) -> Optional[str]:
    """Get the negotiated protocol, if any.

    Args:
        writer (StreamWriter): The writer.

    Returns:
        Optional[str]: The negotiated protocol, if any.
    """
    ssl_object: Optional[ssl.SSLSocket] = writer.get_extra_info('ssl_object')
    if ssl_object is None:
        return None
    negotiated_protocol = ssl_object.selected_alpn_protocol()
    if negotiated_protocol is None:
        negotiated_protocol = ssl_object.selected_npn_protocol()
    return negotiated_protocol



"""Utilities"""

from datetime import datetime
import logging
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Mapping,
    Tuple,
    TypeVar
)

import bareutils.header as header
from bareutils.cookies import encode_cookies

T = TypeVar('T')

LOGGER = logging.getLogger(__name__)


class NullIter(Generic[T]):
    """An iterator containing no items"""

    def __aiter__(self):
        return self

    async def __anext__(self) -> T:
        raise StopAsyncIteration


Cookie = Dict[str, Any]
CookieKey = Tuple[bytes, bytes, bytes]
CookieCache = Dict[CookieKey, Cookie]


def extract_cookies_from_response(
        cookie_cache: CookieCache,
        response: Mapping[str, Any],
        now: datetime
) -> CookieCache:
    """Extract cookies from the response

    Args:
        cookie_cache (CookieCache): The cookie cache
        response (Mapping[str, Any]): The response
        now (datetime): The current time

    Returns:
        CookieCache: The updated cookie cache
    """
    header_cookies = header.set_cookie(response['headers'])
    return extract_cookies(cookie_cache, header_cookies, now)


def extract_cookies(
        cookie_cache: CookieCache,
        header_cookies: Mapping[bytes, List[Dict[str, Any]]],
        now: datetime
) -> CookieCache:
    """Extract cookies from the headers

    Args:
        cookie_cache (CookieCache): The cookie cache
        header_cookies (Mapping[bytes, List[Dict[str, Any]]]): The headers
        now (datetime): The current time

    Returns:
        CookieCache: The updated cookie cache
    """
    current_cookies = {
        key: cookie
        for key, cookie in cookie_cache.items()
        if 'expires' not in cookie or cookie['expires'] > now
    }
    for name, cookies in header_cookies.items():
        for cookie in cookies:
            # Ensure permanant cookies have timestamps.
            if 'max_age' in cookie and 'expires' not in cookie:
                cookie['expires'] = now + cookie['max_age']
            # Don't cache expired cookies
            if 'expires' in cookie and cookie['expires'] < now:
                continue
            path: bytes = cookie.get('path', b'')
            domain: bytes = cookie.get('domain', b'')
            key = (name, domain, path)
            cookie['creation_time'] = now
            cookie['persistant'] = 'expires' in cookie
            current_cookies[key] = cookie
    return current_cookies


def gather_cookies(
        cookie_cache: CookieCache,
        request_scheme: bytes,
        request_domain: bytes,
        request_path: bytes,
        now: datetime
) -> bytes:
    """Gather the cookies from the cookie cache.

    Args:
        cookie_cache (CookieCache): The cookie cache
        request_scheme (bytes): The request scheme
        request_domain (bytes): The request domain
        request_path (bytes): The request path
        now (datetime): The current time

    Returns:
        bytes: The cookie header content
    """
    cookies: Dict[bytes, Cookie] = {}
    for key, cookie in cookie_cache.items():
        if 'expires' in cookie and cookie['expires'] < now:
            del cookie_cache[key]
            continue

        if cookie.get('secure', False) and request_scheme != b'https':
            continue

        domain = cookie.get('domain', b'')
        if domain and not request_domain.endswith(domain):
            continue

        path = cookie.get('path', b'')
        if path and not request_path.startswith(path):
            continue

        name = cookie['name']
        current_cookie = cookies.get(name)
        if current_cookie:
            # Prefer longer domain
            if domain and ('domain' not in cookie or len(domain) < len(cookie['domain'])):
                continue
            # Prefer longer path
            if path and ('path' not in cookie or len(path) < len(cookie['path'])):
                continue

            # Prefer earlier cookies
            if current_cookie['creation_time'] < cookie['creation_time']:
                continue

        cookies[name] = cookie
    for cookie in cookies.values():
        cookie['last_access_time'] = now

    return encode_cookies(
        {
            cookie['name']: [cookie['value']]
            for cookie in cookies.values()
        }
    )



"""bareclient"""

from .client import HttpClient
from .session import HttpSession
from .unbound_session import HttpUnboundSession
from .helpers import (
    request,
    get,
    get_text,
    get_json,
    post,
    post_text,
    post_json
)
from .constants import DEFAULT_PROTOCOLS, DEFAULT_DECOMPRESSORS
from .ssl_contexts import (
    create_ssl_context,
    create_ssl_context_with_cert_chain,
    DEFAULT_CIPHERS,
    DEFAULT_OPTIONS
)

__all__ = [
    'HttpClient',
    'HttpSession',
    'HttpUnboundSession',
    'request',
    'get',
    'get_text',
    'get_json',
    'post',
    'post_text',
    'post_json',
    'create_ssl_context',
    'create_ssl_context_with_cert_chain',
    'DEFAULT_PROTOCOLS',
    'DEFAULT_CIPHERS',
    'DEFAULT_OPTIONS',
    'DEFAULT_DECOMPRESSORS'
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bareclient', 'bareclient.acgi']

package_data = \
{'': ['*']}

install_requires = \
['baretypes>=3.5,<4.0',
 'bareutils>=3.7,<4.0',
 'h11>=0.12.0,<0.13.0',
 'h2>=4.0,<5.0']

setup_kwargs = {
    'name': 'bareclient',
    'version': '4.6.0',
    'description': 'A lightweight asyncio HTTP client',
    'long_description': '# bareClient\n\nA simple asyncio http Python client package supporting HTTP versions 1.0, 1.1\nand 2 (read the [docs](https://rob-blackbourn.github.io/bareClient/)).\n\nThis is the client companion to the ASGI server side web framework\n[bareASGI](https://github.com/rob-blackbourn/bareASGI) and follows the same\n"bare" approach. It makes little attempt to provide any helpful features which\nmight do unnecessary work, providing a foundation for whatever feature set is\nrequired.\n\nIt was written to allow a web server which had negotiated the HTTP/2 protocol\nfor make outgoing HTTP/2 calls. This increases performance and simplifies proxy\nconfiguration in a micro-service architecture.\n\n## Installation\n\nThe package can be installed with pip.\n\n```bash\npip install bareclient\n```\n\nThis is a Python3.7 and later package.\n\nIt has dependencies on:\n\n* [bareTypes](https://github.com/rob-blackbourn/bareTypes)\n* [bareUtils](https://github.com/rob-blackbourn/bareUtils)\n* [h11](https://github.com/python-hyper/h11)\n* [h2](https://github.com/python-hyper/hyper-h2)\n\n## Usage\n\nThe basic usage is to create an `HttpClient`.\n\n```python\nimport asyncio\nfrom typing import List, Optional\nfrom bareclient import HttpClient\n\nasync def main(url: str) -> None:\n    async with HttpClient(url, method=\'GET\') as response:\n        if response[\'status_code\'] == 200 and response[\'more_body\']:\n            async for part in response[\'body\']:\n                print(part)\n\nasyncio.run(main(\'https://docs.python.org/3/library/cgi.html\'))\n```\n\nThere is also an `HttpSession` for maintaining session cookies.\n\n```python\n"""Simple GET"""\n\nimport asyncio\nimport json\n\nfrom bareutils import text_reader\nimport bareutils.header as header\nimport bareutils.response_code as response_code\nfrom bareclient import HttpSession\n\n\nasync def main() -> None:\n    """Session example"""\n\n    session = HttpSession(\'https://jsonplaceholder.typicode.com\')\n    async with session.request(\'/users/1/posts\', method=\'GET\') as response:\n        # We expect a session cookie to be sent on the initial request.\n        set_cookie = header.find(b\'set-cookie\', response[\'headers\'])\n        print("Session cookie!" if set_cookie else "No session cookie")\n\n        if not response_code.is_successful(response[\'status_code\']):\n            raise Exception("Failed to get posts")\n\n        posts = json.loads(await text_reader(response[\'body\']))\n        print(f\'We received {len(posts)} posts\')\n\n        for post in posts:\n            path = f\'/posts/{post["id"]}/comments\'\n            print(f\'Requesting comments from "{path}""\')\n            async with session.request(path, method=\'GET\') as response:\n                # As we were sent the session cookie we do not expect to receive\n                # another one, until this one has expired.\n                set_cookie = header.find(b\'set-cookie\', response[\'headers\'])\n                print("Session cookie!" if set_cookie else "No session cookie")\n\n                if not response_code.is_successful(response[\'status_code\']):\n                    raise Exception("Failed to get comments")\n\n                comments = json.loads(await text_reader(response[\'body\']))\n                print(f\'We received {len(comments)} comments\')\n\n\nasyncio.run(main())\n```\n\nFinally there is a single helper function to get json.\n\n```python\nimport asyncio\n\nfrom bareclient import get_json\n\n\nasync def main(url: str) -> None:\n    """Get some JSON"""\n    obj = await get_json(url, headers=[(b\'accept-encoding\', b\'gzip\')])\n    print(obj)\n\n\nasyncio.run(main(\'https://jsonplaceholder.typicode.com/todos/1\'))\n```\n',
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rob-blackbourn/bareClient',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Connections"""

import asyncio
from asyncio import AbstractEventLoop
from ssl import SSLContext
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Iterable,
    Mapping,
    Optional,
    Union
)
import urllib.parse
from urllib.error import URLError

from ..ssl_contexts import create_ssl_context

from .utils import (
    get_port,
    get_negotiated_protocol
)
from .http_protocol import HttpProtocol
from .h11_protocol import H11Protocol
from .h2_protocol import H2Protocol

SendCallable = Callable[[Mapping[str, Any]], Coroutine[Any, Any, None]]
ReceiveCallable = Callable[[], Awaitable[Mapping[str, Any]]]
Application = Callable[
    [ReceiveCallable, SendCallable],
    Coroutine[Any, Any, Mapping[str, Any]]
]


async def connect(
        url: urllib.parse.ParseResult,
        application: Application,
        cafile: Optional[str],
        capath: Optional[str],
        cadata: Optional[str],
        ssl_context: Optional[SSLContext],
        loop: Optional[AbstractEventLoop],
        h11_bufsiz: int,
        protocols: Iterable[str],
        ciphers: Iterable[str],
        options: Iterable[int],
        connect_timeout: Optional[Union[int, float]]
) -> Mapping[str, Any]:
    """Connect to the web server and run the application

    Args:
        url (urllib.parse.ParseResult): The url
        application (Application): The application to run
        cafile (Optional[str]): The path to a file of concatenated CA
            certificates in PEM format.
        capath (Optional[str]): The path to a directory containing
            several CA certificates in PEM format.
        cadata (Optional[str]): Either an ASCII string of one or more
            PEM-encoded certificates or a bytes-like object of DER-encoded
            certificates.
        ssl_context (Optional[SSLContext]): An ssl context to be
            used instead of generating one from the certificates.
        loop (Optional[AbstractEventLoop]): The optional asyncio event
            loop.
        h11_bufsiz (int): The HTTP/1.1 buffer size.
        protocols (Iterable[str]): The supported protocols.
        ciphers (Iterable[str]): The supported ciphers.
        options (Iterable[int]): The ssl.SSLContext.options.
        connect_timeout (Optional[Union[int, float]]): The connection timeout.

    Raises:
        URLError: Raised for an invalid url
        asyncio.TimeoutError: Raised for a connection timeout.

    Returns:
        Mapping[str, Any]: The response message.
    """
    if ssl_context is None and url.scheme == 'https':
        ssl_context = create_ssl_context(
            cafile,
            capath,
            cadata,
            protocols=protocols,
            ciphers=ciphers,
            options=options
        )

    hostname = url.hostname
    if hostname is None:
        raise URLError('unspecified hostname')
    port = get_port(url)
    if port is None:
        raise URLError('unspecified port')

    future = asyncio.open_connection(
        hostname,
        port,
        loop=loop,
        ssl=ssl_context
    )
    reader, writer = await asyncio.wait_for(future, timeout=connect_timeout)

    negotiated_protocol = get_negotiated_protocol(
        writer
    ) if ssl_context else None

    if negotiated_protocol == 'h2':
        http_protocol: HttpProtocol = H2Protocol(reader, writer)
    else:
        http_protocol = H11Protocol(reader, writer, h11_bufsiz)

    return await application(http_protocol.receive, http_protocol.send)



"""An HTTP session"""

from __future__ import annotations

from asyncio import AbstractEventLoop
from datetime import datetime, timezone
from typing import (
    Any,
    AsyncContextManager,
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Type
)
from urllib.parse import urlparse

from baretypes import Header, Content
from bareutils.compression import Decompressor

from bareclient.client import DEFAULT_DECOMPRESSORS, HttpClient
from bareclient.utils import (
    Cookie,
    extract_cookies,
    extract_cookies_from_response,
    gather_cookies
)
from bareclient.acgi.utils import get_authority

from .constants import DEFAULT_PROTOCOLS
from .ssl_contexts import DEFAULT_CIPHERS, DEFAULT_OPTIONS

HttpClientFactory = Callable[
    [],
    AsyncContextManager[Mapping[str, Any]]
]


class HttpUnboundSessionInstance:
    """An HTTP unbound session instance"""

    def __init__(
            self,
            client: HttpClient,
            update_session: Callable[[Mapping[str, Any]], None]
    ) -> None:
        """Initialise an HTTP unbound session instance.

        Args:
            client (HttpClient): The HTTP client
            update_session (Callable[[Mapping[str, Any]], None]): A function to
                update the session.
        """
        self.client = client
        self.update_session = update_session

    async def __aenter__(self) -> Mapping[str, Any]:
        response = await self.client.__aenter__()
        self.update_session(response)
        return response

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.client.__aexit__(exc_type, exc_val, exc_tb)


class HttpUnboundSession:
    """An HTTP Unbound Session"""

    def __init__(
            self,
            *,
            headers: Optional[List[Header]] = None,
            cookies: Optional[Mapping[bytes, List[Cookie]]] = None,
            loop: Optional[AbstractEventLoop] = None,
            h11_bufsiz: int = 8096,
            cafile: Optional[str] = None,
            capath: Optional[str] = None,
            cadata: Optional[str] = None,
            decompressors: Optional[Mapping[bytes, Type[Decompressor]]] = None,
            protocols: Iterable[str] = DEFAULT_PROTOCOLS,
            ciphers: Iterable[str] = DEFAULT_CIPHERS,
            options: Iterable[int] = DEFAULT_OPTIONS
    ) -> None:
        """Initialise an HTTP session

        The following makes a get request from a session:

        ```python
        import asyncio
        from bareclient import HttpClient


        async def main(url: str) -> None:
            session =  HttpUnboundSession()
            async with session.request(url, method='GET') as response:
                print(response)
                if response['status_code'] == 200 and response['more_body']:
                    async for part in response['body']:
                        print(part)

        asyncio.run(main('https://docs.python.org/3/library/cgi.html'))
        ```        

        Args:
            headers (Optional[List[Header]], optional): The headers. Defaults to
                None.
            cookies (Optional[Mapping[bytes, List[Cookie]]], optional): The
                cookies. Defaults to None.
            loop (Optional[AbstractEventLoop], optional): The asyncio event
                loop. Defaults to None.
            h11_bufsiz (int, optional): The HTTP/1 buffer size. Defaults to 8096.
            loop (Optional[AbstractEventLoop], optional): The optional asyncio
                event loop.. Defaults to None.
            cafile (Optional[str], optional): The path to a file of concatenated
                CA certificates in PEM format. Defaults to None.
            capath (Optional[str], optional): The path to a directory containing
                several CA certificates in PEM format. Defaults to None.
            cadata (Optional[str], optional): Either an ASCII string of one or
                more PEM-encoded certificates or a bytes-like object of
                DER-encoded certificates. Defaults to None.
            decompressors (Optional[Mapping[bytes, Type[Decompressor]]], optional):
                The decompressors. Defaults to None.
            protocols (Iterable[str], optional): The supported protocols.
                Defaults to DEFAULT_PROTOCOLS.
            ciphers (Iterable[str], optional): The supported ciphers.
                Defaults to DEFAULT_CIPHERS.
            options (Iterable[int], optional): The SSLContext options.
                Defaults to DEFAULT_OPTIONS.
        """
        self.headers = headers or []
        self.loop = loop
        self.h11_bufsiz = h11_bufsiz
        self.cafile = cafile
        self.capath = capath
        self.cadata = cadata
        self.decompressors = decompressors or DEFAULT_DECOMPRESSORS
        self.protocols = protocols
        self.ciphers = ciphers
        self.options = options
        self.cookies = extract_cookies({}, cookies or {}, datetime.utcnow())

    def request(
            self,
            url: str,
            *,
            method: str = 'GET',
            headers: Optional[List[Header]] = None,
            content: Optional[Content] = None
    ) -> HttpUnboundSessionInstance:
        """Make an HTTP request

        Args:
            url (str): The url
            method (str, optional): The HTTP method, defaults to 'GET'. Defaults
                to 'GET'.
            headers (Optional[List[Header]], optional): Optional headers.
                Defaults to None.
            content (Optional[Content], optional): Optional content, defaults to
                None. Defaults to None.

        Returns:
            HttpSessionInstance: A context instance yielding the response and body
        """
        combined_headers = self.headers
        if headers:
            combined_headers = combined_headers + headers

        parsed_url = urlparse(url)
        scheme = parsed_url.scheme.encode('ascii')
        domain = get_authority(parsed_url).encode('ascii')

        cookies = self._gather_cookies(
            scheme,
            domain,
            parsed_url.path.encode('ascii')
        )
        if cookies:
            combined_headers.append(
                (b'cookie', cookies)
            )

        client = HttpClient(
            url,
            method=method,
            headers=combined_headers,
            content=content,
            loop=self.loop,
            h11_bufsiz=self.h11_bufsiz,
            cafile=self.cafile,
            capath=self.capath,
            cadata=self.cadata,
            decompressors=self.decompressors,
            protocols=self.protocols,
            ciphers=self.ciphers,
            options=self.options
        )

        return HttpUnboundSessionInstance(client, self._extract_cookies)

    def _extract_cookies(self, response: Mapping[str, Any]) -> None:
        now = datetime.now().astimezone(timezone.utc)
        self.cookies = extract_cookies_from_response(
            self.cookies, response, now
        )

    def _gather_cookies(
            self,
            scheme: bytes,
            domain: bytes,
            path: bytes
    ) -> bytes:
        now = datetime.now().astimezone(timezone.utc)
        return gather_cookies(
            self.cookies,
            scheme,
            domain,
            path,
            now
        )



"""Utilities"""

import asyncio
from typing import Generic, Optional, TypeVar

T = TypeVar('T')


class MessageEvent(asyncio.Event, Generic[T]):
    """An event taking a message when set"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message: Optional[T] = None

    def set_with_message(self, message: Optional[T]) -> None:
        """Set the event with a value

        Args:
            message (Optional[T]): The value to set
        """
        self.message = message
        super().set()

    async def wait_with_message(self) -> Optional[T]:
        """Wait for the event to be set, returning the message dictionary

        Returns:
            Optional[T]: value the event was set with
        """
        await super().wait()
        message, self.message = self.message, None
        return message


class ResetEvent(asyncio.Event):
    """An event which automatically clears after being set"""

    def set(self) -> None:
        """Sets then clears the event"""
        super().set()
        super().clear()



"""Constants"""

import os
from typing import Mapping

from bareutils.compression import (
    make_gzip_decompressobj,
    make_deflate_decompressobj,
    DecompressorFactory
)
import pkg_resources

DIST_VERSION = pkg_resources.get_distribution('bareclient').version

Decompressors = Mapping[bytes, DecompressorFactory]

DEFAULT_DECOMPRESSORS: Decompressors = {
    b'gzip': make_gzip_decompressobj,
    b'deflate': make_deflate_decompressobj
}

SYSNAME, HOST, RELEASE, OS_VERSION, MACHINE = os.uname()


USER_AGENT = f'bareClient/{DIST_VERSION} ({SYSNAME}; {RELEASE}; {MACHINE})'.encode(
    'ascii'
)

DEFAULT_PROTOCOLS = ("h2", "http/1.1")
